@@ -13,13 +13,15 @@ $TRANSDEPTH = 0;
 
 our $VERSION = '0.01';
 
-if (my $pattern = $ENV{JIFTY_DBQUERY_CALLER}) {
+if ( my $pattern = $ENV{JIFTY_DBQUERY_CALLER} ) {
     require Hook::LexWrap;
-    Hook::LexWrap::wrap('Jifty::DBI::Handle::simple_query', pre => sub {
-        return unless $_[1] =~ m/$pattern/;
-        warn $_[1].'   '.join(',', @_[2..$#_])."\n";
-        Carp::cluck;
-    });
+    Hook::LexWrap::wrap(
+        'Jifty::DBI::Handle::simple_query',
+        pre => sub {
+            return unless $_[1] =~ m/$pattern/;
+            Carp::cluck($_[1] . '   ' . CORE::join( ',', @_[ 2 .. $#_ ] ));
+        }
+    );
 }
 
 =head1 NAME
@@ -71,6 +73,10 @@ If you created the handle with
 and there is a Jifty::DBI::Handle::(Driver) subclass for the driver you have chosen,
 the handle will be automatically "upgraded" into that subclass.
 
+If there is an error, an exception will be thrown. If a connection has already
+been established and is still active, C<undef> will be returned (which is not
+an error). Otherwise, if a new connection is made, a true value will be returned.
+
 =cut
 
 sub connect {
@@ -85,6 +91,7 @@ sub connect {
         user       => undef,
         password   => undef,
         requiressl => undef,
+        extra      => {},
         @_
     );
 
@@ -106,8 +113,8 @@ sub connect {
     if ( ( !$self->dbh ) || ( !$self->dbh->ping ) || ( $self->dsn ne $dsn ) )
     {
         my $handle
-            = DBI->connect( $self->dsn, $args{'user'}, $args{'password'} )
-            || Carp::croak "Connect Failed $DBI::errstr\n";
+            = DBI->connect( $self->dsn, $args{'user'}, $args{'password'}, $args{'extra'} )
+            || Carp::croak "Connection failed: $DBI::errstr\n";
 
 #databases do case conversion on the name of columns returned.
 #actually, some databases just ignore case. this smashes it to something consistent
@@ -188,12 +195,16 @@ sub build_dsn {
         @_
     );
 
-
     my $driver = delete $args{'driver'};
     $args{'dbname'} ||= delete $args{'database'};
 
-    $self->{'dsn'} =
-    "dbi:$driver:" . join(';', map { $_ ."=".$args{$_} } grep { defined $args{$_} } keys %args);
+    delete $args{'user'};
+    delete $args{'password'};
+    delete $args{'extra'};
+
+    $self->{'dsn'} = "dbi:$driver:"
+        . CORE::join( ';',
+        map { $_ . "=" . $args{$_} } grep { defined $args{$_} } keys %args );
 }
 
 =head2 dsn
@@ -231,6 +242,21 @@ sub print_error {
     return $self->dbh->{PrintError};
 }
 
+=head2 log MESSAGE
+
+Takes a single argument, a message to log.
+
+Currently prints that message to STDERR
+
+=cut
+
+sub log {
+    my $self = shift;
+    my $msg  = shift;
+    warn $msg . "\n";
+
+}
+
 =head2 log_sql_statements BOOL
 
 Takes a boolean argument. If the boolean is true, it will log all SQL
@@ -249,7 +275,31 @@ sub log_sql_statements {
     return ( $self->{'_dologsql'} );
 }
 
-=head2 _log_sql_statement STATEMENT DURATION
+=head2 log_sql_hook NAME [, CODE]
+
+Used in instrumenting the SQL logging. You can use this to, for example, get a
+stack trace for each query (so you can find out where the query is being made).
+The name is required so that multiple hooks can be installed, and inspected, by
+name.
+
+The coderef is run in scalar context and (currently) receives no arguments.
+
+If you don't pass CODE in, then the coderef currently assigned under
+NAME is returned.
+
+=cut
+
+sub log_sql_hook {
+    my $self = shift;
+    my $name = shift;
+
+    if (@_) {
+        $self->{'_logsqlhooks'}{$name} = shift;
+    }
+    return ( $self->{'_logsqlhooks'}{$name} );
+}
+
+=head2 _log_sql_statement STATEMENT DURATION BINDINGS
 
 add an SQL statement to our query log
 
@@ -260,9 +310,15 @@ sub _log_sql_statement {
     my $statement = shift;
     my $duration  = shift;
     my @bind      = @_;
-    push @{ $self->{'StatementLog'} },
-        ( [ Time::HiRes::time(), $statement, [@bind], $duration ] );
 
+    my %results;
+    my @log = (Time::HiRes::time(), $statement, [@bind], $duration, \%results);
+
+    while (my ($name, $code) = each %{ $self->{'_logsqlhooks'} || {} }) {
+        $results{$name} = $code->(@log);
+    }
+
+    push @{ $self->{'StatementLog'} }, \@log;
 }
 
 =head2 clear_sql_statement_log
@@ -278,9 +334,12 @@ sub clear_sql_statement_log {
 
 =head2 sql_statement_log
 
-Returns the current SQL statement log as an array of arrays. Each entry is a list of 
+Returns the current SQL statement log as an array of arrays. Each entry is a list of:
 
-(Time, Statement, [Bindings], Duration)
+(Time, Statement, [Bindings], Duration, {HookResults})
+
+Bindings is an arrayref of the values of any placeholders. HookResults is a
+hashref keyed by hook name.
 
 =cut
 
@@ -346,7 +405,7 @@ sub delete {
 
     my @bind  = ();
     my $where = 'WHERE ';
-    while (my $key = shift @pairs) {
+    while ( my $key = shift @pairs ) {
         $where .= $key . "=?" . " AND ";
         push( @bind, shift(@pairs) );
     }
@@ -364,7 +423,7 @@ Takes a table name and a set of key-value pairs in an array. splits the key valu
 
 sub insert {
     my ( $self, $table, @pairs ) = @_;
-    my ( @cols, @vals,  @bind );
+    my ( @cols, @vals, @bind );
 
 #my %seen; #only the *first* value is used - allows drivers to specify default
     while ( my $key = shift @pairs ) {
@@ -376,7 +435,8 @@ sub insert {
         push @bind, $value;
     }
 
-    my $query_string = "INSERT INTO $table ("
+    my $query_string
+        = "INSERT INTO $table ("
         . CORE::join( ", ", @cols )
         . ") VALUES " . "("
         . CORE::join( ", ", @vals ) . ")";
@@ -407,7 +467,7 @@ sub update_record_value {
         @_
     );
 
-    return 1 unless grep {defined} values %{$args{primary_keys}};
+    return 1 unless grep {defined} values %{ $args{primary_keys} };
 
     my @bind  = ();
     my $query = 'UPDATE ' . $args{'table'} . ' ';
@@ -461,6 +521,7 @@ Execute the SQL string specified in QUERY_STRING
 
 =cut
 
+our $retry_simple_query = 1;
 sub simple_query {
     my $self         = shift;
     my $query_string = shift;
@@ -469,17 +530,16 @@ sub simple_query {
 
     my $sth = $self->dbh->prepare($query_string);
     unless ($sth) {
+        my $message = "$self couldn't prepare the query '$query_string': "
+                    . $self->dbh->errstr;
         if ($DEBUG) {
-            die "$self couldn't prepare the query '$query_string'"
-                . $self->dbh->errstr . "\n";
+            die "$message\n";
         } else {
-            warn "$self couldn't prepare the query '$query_string'"
-                . $self->dbh->errstr . "\n";
+            warn "$message\n";
             my $ret = Class::ReturnValue->new();
             $ret->as_error(
                 errno   => '-1',
-                message => "Couldn't prepare the query '$query_string'."
-                    . $self->dbh->errstr,
+                message => $message,
                 do_backtrace => undef
             );
             return ( $ret->return_value );
@@ -510,6 +570,18 @@ sub simple_query {
     {
         no warnings 'uninitialized';    # undef in bind_values makes DBI sad
         eval { $executed = $sth->execute(@bind_values) };
+
+        # try to ping and reconnect, if the DB connection failed
+        if (($@ or not $executed) and !$self->dbh->ping) {
+            $self->dbh(undef); # don't try pinging again, just connect
+            $self->connect; 
+
+            # Need to call ourselves, to create a new sth from the new dbh
+            if ($retry_simple_query) {
+                local $retry_simple_query = 0;
+                return $self->simple_query($query_string, @_);
+            }
+        }
     }
     if ( $self->log_sql_statements ) {
         $self->_log_sql_statement( $query_string,
@@ -518,20 +590,21 @@ sub simple_query {
     }
 
     if ( $@ or !$executed ) {
-        if ($DEBUG) {
-            die "$self couldn't execute the query '$query_string'"
-                . $self->dbh->errstr . "\n";
+        my $message = "$self couldn't execute the query '$query_string': "
+                        . ($self->dbh->errstr || $@);
 
+        if ($DEBUG) {
+            die "$message\n";
         } else {
-            # XXX: This warn doesn't show up because we mask logging in Jifty::Test::END.
-            # and it usually fails because the test server is still running.
-            warn "$self couldn't execute the query '$query_string'";
+
+ # XXX: This warn doesn't show up because we mask logging in Jifty::Test::END.
+ # and it usually fails because the test server is still running.
+            warn "$message\n";
 
             my $ret = Class::ReturnValue->new();
             $ret->as_error(
                 errno   => '-1',
-                message => "Couldn't execute the query '$query_string'"
-                    . $self->dbh->errstr,
+                message => $message,
                 do_backtrace => undef
             );
             return ( $ret->return_value );
@@ -583,7 +656,7 @@ Returns the database's version.
 
 If argument C<short> is true returns short variant, in other
 case returns whatever database handle/driver returns. By default
-returns short version, e.g. '4.1.23' or '8.0-rc4'.
+returns short version, e.g. C<4.1.23> or C<8.0-rc4>.
 
 Returns empty string on error or if database couldn't return version.
 
@@ -609,7 +682,7 @@ sub database_version {
         my $ver = '';
         $ver = ( $sth->fetchrow_arrayref->[0] || '' ) if $sth;
         $ver =~ /(\d+(?:\.\d+)*(?:-[a-z0-9]+)?)/i;
-        $self->{'database_version'}       = $ver;
+        $self->{'database_version'} = $ver;
         $self->{'database_version_short'} = $1 || $ver;
 
         $self->raise_error($re);
@@ -648,10 +721,11 @@ sub _case_insensitivity_valid {
     my $value    = shift;
 
     return $value ne ''
-      && $value   ne "''"
-      && ( $operator !~ /IS/ && $value !~ /^null$/i )
-      # don't downcase integer values
-      && $value !~ /^['"]?\d+['"]?$/;
+        && $value ne "''"
+        && ( $operator =~ /^(?:(?:NOT )?LIKE|!?=|IN)$/i )
+
+        # don't downcase integer values
+        && $value !~ /^['"]?\d+['"]?$/;
 }
 
 sub _make_clause_case_insensitive {
@@ -660,13 +734,40 @@ sub _make_clause_case_insensitive {
     my $operator = shift;
     my $value    = shift;
 
-    if ($self->_case_insensitivity_valid($column, $operator, $value)) {
-      $column = "lower($column)";
-      $value  = "lower($value)";
+    if ( $self->_case_insensitivity_valid( $column, $operator, $value ) ) {
+        $column = "lower($column)";
+        if ( ref $value eq 'ARRAY' ) {
+            map { $_ = "lower($_)" } @{$value};
+        } else {
+            $value = "lower($value)";
+        }
     }
     return ( $column, $operator, $value );
 }
 
+=head2 quote_value VALUE
+
+Calls the database's L<DBD/quote> method and returns the result.
+Additionally, turns on perl's utf8 flag if the returned content is
+UTF8.
+
+=cut
+
+sub quote_value {
+    my $self = shift;
+    my ($value) = @_;
+    my $tmp = $self->dbh->quote($value);
+
+    # Accomodate DBI drivers that don't understand UTF8
+    if ( $] >= 5.007 ) {
+        require Encode;
+        if ( Encode::is_utf8($tmp) ) {
+            Encode::_utf8_on($tmp);
+        }
+    }
+    return $tmp;
+}
+
 =head2 begin_transaction
 
 Tells Jifty::DBI to begin a new SQL transaction. This will
@@ -678,12 +779,18 @@ Emulates nested transactions, by keeping a transaction stack depth.
 
 sub begin_transaction {
     my $self = shift;
-    $TRANSDEPTH++;
-    if ( $TRANSDEPTH > 1 ) {
-        return ($TRANSDEPTH);
-    } else {
-        return ( $self->dbh->begin_work );
+
+    if ( $TRANSDEPTH > 0 ) {
+        # We're inside a transaction.
+        $TRANSDEPTH++;
+        return $TRANSDEPTH;
     }
+
+    my $rv = $self->dbh->begin_work;
+    if ($rv) {
+        $TRANSDEPTH++;
+    }
+    return $rv;
 }
 
 =head2 commit
@@ -699,13 +806,18 @@ sub commit {
         Carp::confess(
             "Attempted to commit a transaction with none in progress");
     }
-    $TRANSDEPTH--;
 
-    if ( $TRANSDEPTH == 0 ) {
-        return ( $self->dbh->commit );
-    } else {    #we're inside a transaction
-        return ($TRANSDEPTH);
+    if ($TRANSDEPTH > 1) {
+        # We're inside a nested transaction.
+        $TRANSDEPTH--;
+        return $TRANSDEPTH;
     }
+
+    my $rv = $self->dbh->commit;
+    if ($rv) {
+        $TRANSDEPTH--;
+    }
+    return $rv;
 }
 
 =head2 rollback [FORCE]
@@ -730,16 +842,26 @@ sub rollback {
 #unless ($TRANSDEPTH) {Carp::confess("Attempted to rollback a transaction with none in progress")};
     if ($force) {
         $TRANSDEPTH = 0;
+
         return ( $dbh->rollback );
     }
 
-    $TRANSDEPTH-- if ( $TRANSDEPTH >= 1 );
-    if ( $TRANSDEPTH == 0 ) {
-        return ( $dbh->rollback );
-    } else {    #we're inside a transaction
-        return ($TRANSDEPTH);
+    if ($TRANSDEPTH == 0) {
+        # We're not actually in a transaction.
+        return 1;
     }
 
+    if ($TRANSDEPTH > 1) {
+        # We're inside a nested transaction.
+        $TRANSDEPTH--;
+        return $TRANSDEPTH;
+    }
+
+    my $rv = $dbh->rollback;
+    if ($rv) {
+        $TRANSDEPTH--;
+    }
+    return $rv;
 }
 
 =head2 force_rollback
@@ -799,229 +921,365 @@ L<Jifty::DBI::Collection> object'.
 
 This performs the join.
 
-
 =cut
 
 sub join {
 
     my $self = shift;
     my %args = (
-        collection => undef,
-        type       => 'normal',
-        column1    => 'main',
-        alias1     => undef,
-        table2     => undef,
-        column2    => undef,
-        alias2     => undef,
-        expression => undef,
-        operator   => '=',
+        collection  => undef,
+        type        => 'normal',
+        alias1      => 'main',
+        column1     => undef,
+        table2      => undef,
+        alias2      => undef,
+        column2     => undef,
+        expression  => undef,
+        operator    => '=',
+        is_distinct => 0,
         @_
     );
 
-    my $string;
-
     my $alias;
 
-    #If we're handed in an alias2, we need to go remove it from the
-    # Aliases array.  Basically, if anyone generates an alias and then
-    # tries to use it in a join later, we want to be smart about creating
-    # joins, so we need to go rip it out of the old aliases table and drop
-    # it in as an explicit join
-    if ( $args{'alias2'} ) {
-
-        # this code is slow and wasteful, but it's clear.
-        my @aliases = @{ $args{'collection'}->{'aliases'} };
-        my @new_aliases;
-        foreach my $old_alias (@aliases) {
-            if ( $old_alias =~ /^(.*?) ($args{'alias2'})$/ ) {
-                $args{'table2'} = $1;
-                $alias = $2;
-
-            } else {
-                push @new_aliases, $old_alias;
+    # If we're handed in a table2 as a Collection object, make notes
+    # about if the result of the join is still distinct for the
+    # calling collection
+    if ( $args{'table2'}
+        && UNIVERSAL::isa( $args{'table2'}, 'Jifty::DBI::Collection' ) )
+    {
+        my $c = ref $args{'table2'} ? $args{'table2'} : $args{'table2'}->new($args{collection}->_new_collection_args);
+        if ( $args{'operator'} eq '=' ) {
+            my $x = $c->new_item->column( $args{column2} );
+            if ( $x->type eq 'serial' || $x->distinct ) {
+                $args{'is_distinct'} = 1;
             }
         }
+        $args{'class2'} = ref $c;
+        $args{'table2'} = $c->table;
+    }
 
-# If we found an alias, great. let's just pull out the table and alias for the other item
-        unless ($alias) {
+    if ( $args{'alias2'} ) {
+        if ( $args{'collection'}{'joins'}{ $args{alias2} } and lc $args{'collection'}{'joins'}{ $args{alias2} }{type} eq "cross" ) {
+            my $join = $args{'collection'}{'joins'}{ $args{alias2} };
+            $args{'table2'} = $join->{table};
+            $alias = $join->{alias};
+        } else {
 
             # if we can't do that, can we reverse the join and have it work?
-            my $a1 = $args{'alias1'};
-            my $f1 = $args{'column1'};
-            $args{'alias1'}  = $args{'alias2'};
-            $args{'column1'} = $args{'column2'};
-            $args{'alias2'}  = $a1;
-            $args{'column2'} = $f1;
+            @args{qw/alias1 alias2/}   = @args{qw/alias2 alias1/};
+            @args{qw/column1 column2/} = @args{qw/column2 column1/};
 
-            @aliases     = @{ $args{'collection'}->{'aliases'} };
-            @new_aliases = ();
-            foreach my $old_alias (@aliases) {
-                if ( $old_alias =~ /^(.*?) ($args{'alias2'})$/ ) {
-                    $args{'table2'} = $1;
-                    $alias = $2;
+            if ( $args{'collection'}{'joins'}{ $args{alias2} } and lc $args{'collection'}{'joins'}{ $args{alias2} }{type} eq "cross" ) {
+                my $join = $args{'collection'}{'joins'}{ $args{alias2} };
+                $args{'table2'} = $join->{table};
+                $alias = $join->{alias};
+            } else {
 
-                } else {
-                    push @new_aliases, $old_alias;
-                }
+                # Swap back
+                @args{qw/alias1 alias2/}   = @args{qw/alias2 alias1/};
+                @args{qw/column1 column2/} = @args{qw/column2 column1/};
+
+                return $args{'collection'}->limit(
+                    entry_aggregator => 'AND',
+                    @_,
+                    quote_value => 0,
+                    alias       => $args{'alias1'},
+                    column      => $args{'column1'},
+                    value       => $args{'alias2'} . "." . $args{'column2'},
+                );
             }
-
         }
-
-        if ( !$alias || $args{'alias1'} ) {
-            return ( $self->_normal_join(%args) );
-        }
-
-        $args{'collection'}->{'aliases'} = \@new_aliases;
-    }
-
-    else {
+    } else {
         $alias = $args{'collection'}->_get_alias( $args{'table2'} );
-
     }
 
+    my $meta = $args{'collection'}->{'joins'}{$alias} ||= {};
+    $meta->{alias} = $alias;
     if ( $args{'type'} =~ /LEFT/i ) {
-
-        $string = " LEFT JOIN " . $args{'table2'} . " $alias ";
+        $meta->{'alias_string'}
+            = " LEFT JOIN " . $args{'table2'} . " $alias ";
+        $meta->{'type'} = 'LEFT';
 
     } else {
-
-        $string = " JOIN " . $args{'table2'} . " $alias ";
-
+        $meta->{'alias_string'} = " JOIN " . $args{'table2'} . " $alias ";
+        $meta->{'type'}         = 'NORMAL';
     }
+    $meta->{'depends_on'}       = $args{'alias1'};
+    $meta->{'is_distinct'}      = $args{'is_distinct'};
+    $meta->{'class'}            = $args{'class2'} if $args{'class2'};
+    $meta->{'entry_aggregator'} = $args{'entry_aggregator'}
+        if $args{'entry_aggregator'};
 
-    my $criterion;
-    if ( $args{'expression'} ) {
-        $criterion = $args{'expression'};
-    } else {
-        $criterion = $args{'alias1'} . "." . $args{'column1'};
-    }
-
-    $args{'collection'}->{'leftjoins'}{"$alias"}{'alias_string'} = $string;
-    $args{'collection'}->{'leftjoins'}{"$alias"}{'entry_aggregator'}
-        = $args{'entry_aggregator'}
-        if ( $args{'entry_aggregator'} );
-    $args{'collection'}->{'leftjoins'}{"$alias"}{'depends_on'}
-        = $args{'alias1'};
-    $args{'collection'}->{'leftjoins'}{"$alias"}{'criteria'}
-        { 'criterion' . $args{'collection'}->{'criteria_count'}++ }
-        = " $criterion $args{'operator'} $alias.$args{'column2'}";
+    my $criterion = $args{'expression'} || "$args{'alias1'}.$args{'column1'}";
+    $meta->{'criteria'}{'base_criterion'} = [
+        {   column   => $criterion,
+            operator => $args{'operator'},
+            value    => "$alias.$args{'column2'}",
+        }
+    ];
 
     return ($alias);
 }
 
-sub _normal_join {
-
-    my $self = shift;
-    my %args = (
-        collection => undef,
-        type       => 'normal',
-        column1    => undef,
-        alias1     => undef,
-        table2     => undef,
-        column2    => undef,
-        alias2     => undef,
-        operator   => '=',
-        @_
-    );
-
-    my $sb = $args{'collection'};
-
-    if ( $args{'type'} =~ /LEFT/i ) {
-        my $alias = $sb->_get_alias( $args{'table2'} );
-
-        $sb->{'leftjoins'}{"$alias"}{'alias_string'}
-            = " LEFT JOIN $args{'table2'} $alias ";
-
-        $sb->{'leftjoins'}{"$alias"}{'criteria'}{'base_criterion'}
-            = " $args{'alias1'}.$args{'column1'} $args{'operator'} $alias.$args{'column2'}";
-
-        return ($alias);
-    } else {
-        $sb->Jifty::DBI::Collection::limit(
-            entry_aggregator => 'AND',
-            quote_value      => 0,
-            alias            => $args{'alias1'},
-            column           => $args{'column1'},
-            value            => $args{'alias2'} . "." . $args{'column2'},
-            @_
-        );
-    }
-}
-
 # this code is all hacky and evil. but people desperately want _something_ and I'm
 # super tired. refactoring gratefully appreciated.
 
 sub _build_joins {
+    my $self       = shift;
+    my $collection = shift;
+
+    $self->_optimize_joins( collection => $collection );
+
+    my @cross = grep { lc $_->{type} eq "cross" }
+        values %{ $collection->{'joins'} };
+    my $join_clause = ( $collection->table . " main" )
+        . CORE::join( " ", map { $_->{alias_string} } @cross );
+    my %processed = map { $_->{alias} => 1 } @cross;
+    $processed{'main'} = 1;
+
+# get a @list of joins that have not been processed yet, but depend on processed join
+    my $joins = $collection->{'joins'};
+    while ( my @list = grep !$processed{$_}
+        && $processed{ $joins->{$_}{'depends_on'} }, keys %$joins )
+    {
+        foreach my $join (@list) {
+            $processed{$join}++;
+
+            my $meta = $joins->{$join};
+            my $aggregator = $meta->{'entry_aggregator'} || 'AND';
+
+            $join_clause .= $meta->{'alias_string'} . " ON ";
+            my @tmp = map {
+                ref($_)
+                    ? $_->{'column'} . ' '
+                    . $_->{'operator'} . ' '
+                    . $_->{'value'}
+                    : $_
+                }
+                map {
+                ( '(', @$_, ')', $aggregator )
+                } values %{ $meta->{'criteria'} };
+
+            # delete last aggregator
+            pop @tmp;
+            $join_clause .= CORE::join ' ', @tmp;
+        }
+    }
+
+# here we could check if there is recursion in joins by checking that all joins
+# are processed
+    if ( my @not_processed = grep !$processed{$_}, keys %$joins ) {
+        die "Unsatisfied dependency chain in joins @not_processed";
+    }
+    return $join_clause;
+}
+
+sub _optimize_joins {
+    my $self  = shift;
+    my %args  = ( collection => undef, @_ );
+    my $joins = $args{'collection'}->{'joins'};
+
+    my %processed;
+    $processed{$_}++
+        foreach grep {lc $joins->{$_}{'type'} ne 'left'} keys %$joins;
+    $processed{'main'}++;
+
+    my @ordered;
+
+# get a @list of joins that have not been processed yet, but depend on processed join
+# if we are talking about forest then we'll get the second level of the forest,
+# but we should process nodes on this level at the end, so we build FILO ordered list.
+# finally we'll get ordered list with leafes in the beginning and top most nodes at
+# the end.
+    while ( my @list = grep !$processed{$_}
+        && $processed{ $joins->{$_}{'depends_on'} }, keys %$joins )
+    {
+        unshift @ordered, @list;
+        $processed{$_}++ foreach @list;
+    }
+
+    foreach my $join (@ordered) {
+        next
+            if $self->may_be_null(
+            collection => $args{'collection'},
+            alias      => $join
+            );
+
+        $joins->{$join}{'alias_string'} =~ s/^\s*LEFT\s+/ /i;
+        $joins->{$join}{'type'} = 'NORMAL';
+    }
+
+}
+
+=head2 may_be_null
+
+Takes a C<collection> and C<alias> in a hash and returns true if
+restrictions of the query allow NULLs in a table joined with the
+alias, otherwise returns false value which means that you can use
+normal join instead of left for the aliased table.
+
+Works only for queries have been built with
+L<Jifty::DBI::Collection/join> and L<Jifty::DBI::Collection/limit>
+methods, for other cases return true value to avoid fault
+optimizations.
+
+=cut
+
+sub may_be_null {
     my $self = shift;
-    my $sb   = shift;
-    my %seen_aliases;
+    my %args = ( collection => undef, alias => undef, @_ );
 
-    $seen_aliases{'main'} = 1;
+# if we have at least one subclause that is not generic then we should get out
+# of here as we can't parse subclauses
+    return 1
+        if grep $_ ne 'generic_restrictions',
+        keys %{ $args{'collection'}->{'subclauses'} };
 
-    # We don't want to get tripped up on a dependency on a simple alias.
-    foreach my $alias ( @{ $sb->{'aliases'} } ) {
-        if ( $alias =~ /^(.*?)\s+(.*?)$/ ) {
-            $seen_aliases{$2} = 1;
-        }
+    # build full list of generic conditions
+    my @conditions;
+    foreach ( grep @$_, values %{ $args{'collection'}->{'restrictions'} } ) {
+        push @conditions, 'AND' if @conditions;
+        push @conditions, '(', @$_, ')';
     }
 
-    my $join_clause = $sb->table . " main ";
+    # find tables that depends on this alias and add their join conditions
+    foreach my $join ( values %{ $args{'collection'}->{'joins'} } ) {
 
-    my @keys = ( keys %{ $sb->{'leftjoins'} } );
-    my %seen;
+        # left joins on the left side so later we'll get 1 AND x expression
+        # which equal to x, so we just skip it
+        next if $join->{'type'} eq 'LEFT';
+        next unless $join->{'depends_on'} && ($join->{'depends_on'} eq $args{'alias'});
 
-    while ( my $join = shift @keys ) {
-        if ( !$sb->{'leftjoins'}{$join}{'depends_on'}
-            || $seen_aliases{ $sb->{'leftjoins'}{$join}{'depends_on'} } )
-        {
-            $join_clause
-                .= $sb->{'leftjoins'}{$join}{'alias_string'} . " ON ";
+        my @tmp = map { ( '(', @$_, ')', $join->{'entry_aggregator'} ) }
+            values %{ $join->{'criteria'} };
+        pop @tmp;
 
-            my @criteria = values %{ $sb->{'leftjoins'}{$join}{'criteria'} };
-            my $entry_aggregator
-                = $sb->{'leftjoins'}{$join}{'entry_aggregator'} || 'AND';
-            my $criteria = CORE::join( " $entry_aggregator ",
-                map {" ( $_ ) "} @criteria );
+        @conditions = ( '(', @conditions, ')', 'AND', '(', @tmp, ')' );
 
-            $join_clause .= "( " . $criteria . " ) ";
-            $join_clause = "(" . $join_clause . ")";
+    }
+    return 1 unless @conditions;
 
-            $seen_aliases{$join} = 1;
+    # replace conditions with boolean result: 1 - allow nulls, 0 - doesn't
+    foreach ( splice @conditions ) {
+        unless ( ref $_ ) {
+            push @conditions, $_;
+        } elsif ( $_->{'column'} =~ /^\Q$args{'alias'}./ ) {
+
+            # only operator IS allows NULLs in the aliased table
+            push @conditions, lc $_->{'operator'} eq 'is';
+        } elsif ( $_->{'value'} && $_->{'value'} =~ /^\Q$args{'alias'}./ ) {
+
+            # right operand is our alias, such condition don't allow NULLs
+            push @conditions, 0;
         } else {
-            push( @keys, $join );
-            die "Unsatisfied dependency chain in joins @keys"
-                if $seen{"@keys"}++;
+
+            # conditions on other aliases
+            push @conditions, 1;
         }
-
     }
-    return ( CORE::join( ", ", ( $join_clause, @{ $sb->{'aliases'} } ) ) );
 
+    # returns index of closing paren by index of openning paren
+    my $closing_paren = sub {
+        my $i     = shift;
+        my $count = 0;
+        for ( ; $i < @conditions; $i++ ) {
+            if ( $conditions[$i] && $conditions[$i] eq '(' ) {
+                $count++;
+            } elsif ( $conditions[$i] && $conditions[$i] eq ')' ) {
+                $count--;
+            }
+            return $i unless $count;
+        }
+        die "lost in parens";
+    };
+
+    # solve boolean expression we have, an answer is our result
+    my @tmp = ();
+    while ( defined( my $e = shift @conditions ) ) {
+
+        #warn "@tmp >>>$e<<< @conditions";
+        return $e if !@conditions && !@tmp;
+
+        unless ($e) {
+            if ( $conditions[0] eq ')' ) {
+                push @tmp, $e;
+                next;
+            }
+
+            my $aggreg = uc shift @conditions;
+            if ( $aggreg eq 'OR' ) {
+
+                # 0 OR x == x
+                next;
+            } elsif ( $aggreg eq 'AND' ) {
+
+                # 0 AND x == 0
+                my $close_p = $closing_paren->(0);
+                splice @conditions, 0, $close_p + 1, (0);
+            } else {
+                die "lost @tmp >>>$e $aggreg<<< @conditions";
+            }
+        } elsif ( $e eq '1' ) {
+            if ( $conditions[0] eq ')' ) {
+                push @tmp, $e;
+                next;
+            }
+
+            my $aggreg = uc shift @conditions;
+            if ( $aggreg eq 'OR' ) {
+
+                # 1 OR x == 1
+                my $close_p = $closing_paren->(0);
+                splice @conditions, 0, $close_p + 1, (1);
+            } elsif ( $aggreg eq 'AND' ) {
+
+                # 1 AND x == x
+                next;
+            } else {
+                die "lost @tmp >>>$e $aggreg<<< @conditions";
+            }
+        } elsif ( $e eq '(' ) {
+            if ( $conditions[1] eq ')' ) {
+                splice @conditions, 1, 1;
+            } else {
+                push @tmp, $e;
+            }
+        } elsif ( $e eq ')' ) {
+            unshift @conditions, @tmp, $e;
+            @tmp = ();
+        } else {
+            die "lost: @tmp >>>$e<<< @conditions";
+        }
+    }
+    return 1;
 }
 
 =head2 distinct_query STATEMENTREF 
 
 takes an incomplete SQL SELECT statement and massages it to return a DISTINCT result set.
 
-
 =cut
 
 sub distinct_query {
     my $self         = shift;
     my $statementref = shift;
-    my $sb           = shift;
+    my $collection   = shift;
 
     # Prepend select query for DBs which allow DISTINCT on all column types.
-    $$statementref = "SELECT DISTINCT ".$sb->_preload_columns." FROM $$statementref";
+    $$statementref
+        = "SELECT DISTINCT "
+        . $collection->query_columns
+        . " FROM $$statementref";
 
-    $$statementref .= $sb->_group_clause;
-    $$statementref .= $sb->_order_clause;
+    $$statementref .= $collection->_group_clause;
+    $$statementref .= $collection->_order_clause;
 }
 
 =head2 distinct_count STATEMENTREF 
 
 takes an incomplete SQL SELECT statement and massages it to return a DISTINCT result set.
 
-
 =cut
 
 sub distinct_count {
@@ -1033,21 +1291,116 @@ sub distinct_count {
 
 }
 
-=head2 log MESSAGE
+=head2 canonical_true
 
-Takes a single argument, a message to log.
+This returns the canonical true value for this database. For example, in SQLite
+it is 1 but in Postgres it is 't'.
 
-Currently prints that message to STDERR
+The default is 1.
 
 =cut
 
-sub log {
-    my $self = shift;
-    my $msg  = shift;
-    warn $msg . "\n";
+sub canonical_true { 1 }
 
+=head2 canonical_false
+
+This returns the canonical false value for this database. For example, in SQLite
+it is 0 but in Postgres it is 'f'.
+
+The default is 0.
+
+=cut
+
+sub canonical_false { 0 }
+
+=head2 Schema manipulation methods
+
+=head3 rename_column
+
+Rename a column in a table. Takes 'table', 'column' and new name in 'to'.
+
+=cut
+
+sub rename_column {
+    my $self = shift;
+    my %args = (table => undef, column => undef, to => undef, @_);
+# Oracle: since Oracle 9i R2
+# Pg: 7.4 can this and may be earlier
+    return $self->simple_query(
+        "ALTER TABLE $args{'table'} RENAME COLUMN $args{'column'} TO $args{'to'}"
+    );
 }
 
+
+=head3 rename_table
+
+Renames a table in the DB. Takes 'table' and new name of it in 'to'.
+
+=cut
+
+sub rename_table {
+    my $self = shift;
+    my %args = (table => undef, to => undef, @_);
+# mysql has RENAME TABLE, but alter can rename temporary
+# Oracle, Pg, SQLite are ok with this
+    return $self->simple_query("ALTER TABLE $args{'table'} RENAME TO $args{'to'}");
+}
+
+=head2 supported_drivers
+
+Returns a list of the drivers L<Jifty::DBI> supports.
+
+=cut
+
+sub supported_drivers {
+    return qw(
+        SQLite
+        Informix
+        mysql
+        mysqlPP
+        ODBC
+        Oracle
+        Pg
+        Sybase
+    );
+}
+
+=head2 available_drivers
+
+Returns a list of the available drivers based on the presence of C<DBD::*>
+modules.
+
+=cut
+
+sub available_drivers {
+    my $self = shift;
+
+    local $@;
+    return grep { eval "require DBD::" . $_ } $self->supported_drivers;
+}
+
+=head2 is_available_driver
+
+Returns a boolean indicating whether the provided driver is available.
+
+=cut
+
+do {
+    # lazily memoize
+    my $is_available_driver;
+
+    sub is_available_driver {
+        my $self   = shift;
+        my $driver = shift;
+
+        if (!$is_available_driver) {
+            %$is_available_driver = map { $_ => 1 } $self->available_drivers;
+        }
+
+        return $is_available_driver->{$driver};
+    }
+};
+
 =head2 DESTROY
 
 When we get rid of the L<Jifty::DBI::Handle>, we need to disconnect
@@ -1057,7 +1410,17 @@ from the database
 
 sub DESTROY {
     my $self = shift;
-    $self->disconnect;
+
+    # eval in DESTROY can cause $@ issues elsewhere
+    local $@;
+
+    $self->disconnect
+        unless $self->dbh
+            and $self->dbh
+                # We use an eval {} because DESTROY order during
+                # global destruction is not guaranteed -- the dbh may
+                # no longer be tied, which throws an error.
+            and eval { $self->dbh->{InactiveDestroy} };
     delete $DBIHandle{$self};
 }
 
@@ -1073,7 +1436,7 @@ also C<DBI> about setting C<DBI_PROFILE>.
 
 =head1 AUTHOR
 
-Jesse Vincent, jesse@fsck.com
+Jesse Vincent, <jesse@bestpractical.com>
 
 =head1 SEE ALSO
 
